// Ambrose Sterr 01/31/2006
// Finite State Machine Simulator - I/O Handler
// For Formal Languages and the Theory of Computation @ Marlboro College
// Version 0.3
// Changes from Version 0.2:
// -Relaxed order restrictions on transitions, inputs, and tests, and allowed multiple banks of each
// Changes from Version 0.1:
// -Verbose mode added 
//
// Usage: java fsm <filename> [-v]
// [-v] enables tests
import java.io.*;
import java.util.StringTokenizer;
public class fsm
{
	public static void main(String[] args)
	{
// A couple checks before we get started to make sure the argument points to a file I can open and to enable testing
		boolean verbose = false;
		if (args.length == 1)
		{
		}
		else if ((args.length == 2) && (args[1].equals("-v")))
		{
			verbose = true;
		}
		else
		{
			System.out.println("Invalid Arguments.\nEnding...");
			System.exit(0);
		}
		try
		{
			BufferedReader inputReader = new BufferedReader(new FileReader(args[0]));
			parse(inputReader, verbose);
		} catch (FileNotFoundException e)
		{
			System.out.println ("I can't find that file!/nEnding...");
			System.exit(0);
		}
	}
// This portion interprets the file, created the machine, feeds the inputs in, and interprets the responses.
	public static void parse(BufferedReader inputReader, boolean verbose)
	{
// These are the variables the machine needs.
		String accept[] = null;
		String alphabet[] = null;
		String start = null;
		String states[] = null;
		try
		{
			StringTokenizer tokenizer;
			String line = null;
			String id;
// count tells me what phase of the parsing I'm in.  numI numerates the inputs and numT the tests.
			int count = 0;
			int numI = 0;
			int numT = 0;
			machine fsMachine = null;
			while(true)
			{
// Reads the next line of the file.
				line = inputReader.readLine();
//DEBUG:				System.out.println("Got me a line! \""+line+"\"");
// End of file detection.
				if (line == null)
				{
					System.out.println("\nEnd of file.\nEnding...");
					System.exit(0);
				}
				tokenizer = new StringTokenizer(line,",:[] \t\n\r\f");
// This interprets empty lines as comments.
				if (tokenizer.hasMoreTokens())
					id = tokenizer.nextToken();
				else
					id = "#";
//DEBUG:				System.out.println("id = "+id);
// Comments are skipped.
				if (id.startsWith("#"))
				{
				}
// These 4 blocks read the states, the alphabet, the accept states, and the start state.  The order of they're in doesn't matter.  When count = 4, I know I've got them all.
// Unless the file repeats one or more of them.  Then the program will crash.
				else if ((count < 4) && (id.equals("states")))
				{
					states = handleList(tokenizer);
					count++;
				}
				else if ((count < 4) && (id.equals("alphabet")))
				{
					alphabet = handleList(tokenizer);
					count++;
				}
				else if ((count < 4) && (id.equals("accept")))
				{
					accept = handleList(tokenizer);
					count++;
				}
				else if ((count < 4) && (id.equals("start")))
				{
					start = tokenizer.nextToken();
					count++;
				}
// Once I've got the first 4 pieces, I start looking for the transitions.  When I find the transitions header, I create the machine.
				else if ((count >= 4) && (id.equals("transitions")))
				{
					if (fsMachine == null)
						fsMachine = new machine(states, start, accept, alphabet);
					count = 5;
				}
// Once the transition header has been found, sequences are interpreted as transitions and taught to the machine.
				else if ((count == 5) && (id.equals("-")))
				{
					String[] rule = handleList(tokenizer);
					fsMachine.makeRule(rule[0], rule[1], rule[2]);
				}
// When the inputs header is found, sequences are then interpreted as input strings and run, and the results printed.
				else if ((count >= 4) && (id.equals("inputs")))
				{
					if (fsMachine == null)
						fsMachine = new machine(states, start, accept, alphabet);
					count = 6;
				}
				else if ((count == 6) && (id.equals("-")))
				{
					numI++;
					if (fsMachine.run(handleList(tokenizer)))
						System.out.println("Input " + numI + " is accepted.");
					else
						System.out.println("Input " + numI + " is not accepted.");
				}
// If a tests header is found and verbose mode is enabled, sequences are then interpreted as tests.  The expected result is recorded and compared to the machine's return.
// Passes and fails are both reported.
				else if ((count >= 4) && (id.equals("tests")))
				{
					if (fsMachine == null)
						fsMachine = new machine(states, start, accept, alphabet);
					count = 7;
				}
				else if ((count == 7) && (id.equals("-")))
				{
					if (verbose)
					{
						numT++;
						String expected = tokenizer.nextToken();
						System.out.println(expected);
						if (!(fsMachine.run(handleList(tokenizer)) ^ expected.equals("Yes")))
						{
							System.out.println("Test " + numT + " passed.");
						}
						else
							System.out.println("Test " + numT + " failed.");
					}
				}
// If the parser encouters a line starting with ---, it assumes the file contains multiple machines, and so everything is cleared and we start over with whatever follows that line.
				else if (id.startsWith("---"))
				{
					accept = null;
					alphabet = null;
					start = null;
					states = null;
					count = 0;
					numI = 0;
					numT = 0;
					fsMachine = null;
					System.out.println("\nI found another machine to run!\nI'll start that one now.\n");
				}
// If there is a line that doesn't make sense or is out of order, the program quits.
				else
				{
					System.out.println("I don't understand this line \"" + line + "\" or I'm not ready for it.\nEnding...");
					System.exit(0);
				}
			}
// Just in case...
			} catch (IOException e)
		{
			System.out.println ("Ooops! IO Exception during input./nEnding...");
			System.exit(0);
		}
	}
// This method parses lists into string arrays.
	public static String[] handleList(StringTokenizer listTokenizer)
	{
		int size = listTokenizer.countTokens();
		String[] list = new String[size];
		for (int i = 0; i < size; i++)
		{
			list[i] = listTokenizer.nextToken();
		}
		return list;
	}
}